#include <string.h>
#include "system.h"
#include "util.h"
#include "type.h"
#include "network.h"
#include "connect.h"
#include "example.h"
#include "train.h"
#include "act.h"
#include "command.h"
#include "control.h"
#include "object.h"
#include "display.h"

/* Add your initials or something here to customize the version number */
char *VersionExt = "CTK";

flag autoseqNetRunExample(Example E, flag (*tickProc)(Event), flag *correct);


/******************************** Extension Records **************************/

ObjInfo NetExtInfo;
ObjInfo GroupExtInfo;
ObjInfo UnitExtInfo;
ObjInfo BlockExtInfo;
ObjInfo ExSetExtInfo;
ObjInfo ExampleExtInfo;
ObjInfo EventExtInfo;

/* This is what the object viewer displays as the name of the extension. */
static void extensionName(void *ext, char *dest) {
  if (ext) sprintf(dest, "extension");
  else dest[0] = '\0';
}


struct netExt {
  int middleGroupNum, sequenceMidpoint;
};

/* This is how you register fields so they are visible in the object viewer 
   and accessible with setObject and getObject */
static void initNetExtInfo(void) {
  NetExt NX;
  addObj(NetExtInfo, "middleGroupNum", OFFSET(NX, middleGroupNum), TRUE, IntInfo);
  addObj(NetExtInfo, "sequenceMidpoint", OFFSET(NX, sequenceMidpoint), TRUE, IntInfo);
}

/* Use this to build the network extension and set your own default 
   parameters */
flag initNetworkExtension(Network N) {
  FREE(N->ext);

  N->ext = safeCalloc(1, sizeof(struct netExt), 
     "initNetworkExtension:N->ext");
  N->ext->middleGroupNum = 0;
  N->ext->sequenceMidpoint = 0;

  return TCL_OK;
}

/* This is called when the network is reset.  If you also want to reset group
   or other extensions, you should do that from within here. */
flag resetNetworkExtension(Network N) {
  N->ext->middleGroupNum = 0;
  N->ext->sequenceMidpoint = 0;
  return TCL_OK;
}

flag freeNetworkExtension(Network N) {
  FREE(N->ext);
  return TCL_OK;
}


struct groupExt {
  int removeMe;
};

static void initGroupExtInfo(void) {
  /*  GroupExt GE; */
}

flag initGroupExtension(Group G) {
  FREE(G->ext);
  return TCL_OK;
}

flag freeGroupExtension(Group G) {
  FREE(G->ext);
  return TCL_OK;
}


struct unitExt {
  int removeMe;
};

static void initUnitExtInfo(void) {
  /*  UnitExt UE; */
}

flag initUnitExtension(Unit U) {
  FREE(U->ext);
  return TCL_OK;
}

flag freeUnitExtension(Unit U) {
  FREE(U->ext);
  return TCL_OK;
}


struct blockExt {
  int removeMe;
};

static void initBlockExtInfo(void) {
  /* BlockExt BE; */
}

flag initBlockExtension(Block B) {
  FREE(B->ext);
  return TCL_OK;
}

flag freeBlockExtension(Block B) {
  FREE(B->ext);
  return TCL_OK;
}

flag copyBlockExtension(Block from, Block to) {
  if (!from->ext) {
    if (to->ext) {
      freeBlockExtension(to);
      to->ext = NULL;
    }
    return TCL_OK;
  }
  if (!to->ext) initBlockExtension(to);
  to->ext->removeMe = from->ext->removeMe;
  return TCL_OK;
}


struct exSetExt {
  int removeMe;
};

static void initExSetExtInfo(void) {
  /* ExSetExt SE; */
}

flag initExSetExtension(ExampleSet S) {
  FREE(S->ext);
  return TCL_OK;
}

flag freeExSetExtension(ExampleSet S) {
  FREE(S->ext);
  return TCL_OK;
}


struct exampleExt {
  int removeMe;
};

static void initExampleExtInfo(void) {
  /* ExampleExt EE; */
}

flag initExampleExtension(Example E) {
  FREE(E->ext);
  return TCL_OK;
}

flag freeExampleExtension(Example E) {
  FREE(E->ext);
  return TCL_OK;
}


struct eventExt {
  int removeMe;
};

static void initEventExtInfo(void) {
  /* EventExt VE; */
}

flag initEventExtension(Event V) {
  FREE(V->ext);
  return TCL_OK;
}

flag freeEventExtension(Event V) {
  FREE(V->ext);
  return TCL_OK;
}


/********************************* New Tcl Commands **************************/
/* Here is a sample Tcl-callable C stub.  util.c contains the functions
   warning(), result(), and append() for creating Tcl return messages.
   The stub must be registered (see below) before it can be called. */


static int C_initAutoSeq(TCL_CMDARGS) {
  char *usage = "initAutoSeq";
  if (argc == 2 && !strcmp(argv[1], "-h"))
    return commandHelp(argv[0]);
  if (argc != 1) return usageError(argv[0], usage);
  if (!Net) return warning("%s: no current network", argv[0]);
  
  Net->netTrainExample     = autoseqNetRunExample;
  Net->netTestExample      = autoseqNetRunExample;
  Net->netRunExample       = autoseqNetRunExample;

  return TCL_OK;
}

 
/******************************* User Initialization *************************/

/* This gets called while Lens is booting up */
flag userInit(void) {
  NetExtInfo = newObject("Network Extension", sizeof(struct netExt), 1,
			 extensionName, NULL, NULL);
  GroupExtInfo = newObject("Group Extension", sizeof(struct groupExt), 3,
			   extensionName, NULL, NULL);
  UnitExtInfo = newObject("Unit Extension", sizeof(struct unitExt), 5,
			  extensionName, NULL, NULL);
  BlockExtInfo = newObject("Block Extension", sizeof(struct blockExt), 7,
			   extensionName, NULL, NULL);
  ExSetExtInfo = newObject("Example Set Extension", sizeof(struct exSetExt), 3,
			   extensionName, NULL, NULL);
  ExampleExtInfo = newObject("Example Extension", sizeof(struct exampleExt), 5,
			     extensionName, NULL, NULL);
  EventExtInfo = newObject("Event Extension", sizeof(struct eventExt), 7,
			   extensionName, NULL, NULL);
  initNetExtInfo();
  initGroupExtInfo();
  initUnitExtInfo();
  initBlockExtInfo();
  initExSetExtInfo();
  initExampleExtInfo();
  initEventExtInfo();

  /* This creates a blank line in the list of commands produced by "help": */
  registerCommand(NULL, "", "");
  registerCommand(C_initAutoSeq, "initAutoSeq", "run current net as an autosequencer"); 

  return TCL_OK;
}


flag inputSideForward( flag middleCostTime ) {
  int g; Group G;
  for (g = 0; g <= Net->ext->middleGroupNum; g++) {
    G = Net->group[g];
    computeInput(G,  TRUE);
    computeOutput(G, TRUE);
    if( g == Net->ext->middleGroupNum ) {
      if( middleCostTime ) {
	computeCost(G, FALSE);
	computeCostBack(G, FALSE);
	resetOutputDerivCache(G);
      } else {
	resetOutputDerivs(G);
      }
    }
  }

  G = Net->group[Net->numGroups-1];
  computeInput(G,  TRUE);
  computeOutput(G, TRUE);

  for (g = Net->ext->middleGroupNum+1; g < Net->numGroups-1; g++) {
    G = Net->group[g];
    resetOutputs(G);
    resetForwardIntegrators(G);
    computeCost(G, FALSE);
    storeOutputs(G, Net->currentTick);
  }

  return TCL_OK;
}

flag outputSideForwardBackward() {
  int g; Group G;

  for (g = 0; g < Net->ext->middleGroupNum; g++) {
    G = Net->group[g];
    resetOutputs(G);
    resetForwardIntegrators(G);
    storeOutputs(G, Net->currentTick);
  }

  G = Net->group[g];
  storeInputs(G, Net->currentTick);
  storeOutputs(G, Net->currentTick);

  for (g = Net->ext->middleGroupNum+1; g < Net->numGroups; g++) {
    G = Net->group[g];
    computeInput(G,  FALSE);
    computeOutput(G, FALSE);
    computeCost(G,   FALSE);
    computeCostBack(G, FALSE);
    resetOutputDerivCache(G);
  }

  for (g = Net->numGroups - 1; g > Net->ext->middleGroupNum; g--) {
    G = Net->group[g];
    computeOutputBack(G);
    computeInputBack(G);
  }

  return TCL_OK;
}

flag inputSideBackward() {
  int g; Group G;

  for (g = 0; g < Net->ext->middleGroupNum; g++) {
    G = Net->group[g];
    computeCost(G, FALSE);
    computeCostBack(G, FALSE);
    resetOutputDerivCache(G);
  }

  for (g = Net->ext->middleGroupNum; g >= 0; g--) {
    G = Net->group[g];
    computeOutputBack(G);
    computeInputBack(G);
  }

  return TCL_OK;
}


flag autoseqNetRunExample(Example E, flag (*tickProc)(Event), flag *correct) {
  int tick = 0, ticksOnEvent, event = 0, mp = Net->ext->sequenceMidpoint;
  flag eventDone, critReached, training = (tickProc == Net->netTrainTick);
  real timeOnEvent, minTime, maxTime, graceTime;
  Event V;

  if (E->set->loadEvent(V = E->event)) return TCL_ERROR;
  ticksOnEvent = 0;
  maxTime = chooseValue(V->maxTime, E->set->maxTime);
  minTime = chooseValue(V->minTime, E->set->minTime);
  graceTime = chooseValue(V->graceTime, E->set->graceTime);
  Net->inGracePeriod = (graceTime > 0.0);
  
  /* Reset the outputs and integrators if requested. */
  FOR_EACH_GROUP( resetOutputs(G);
    resetForwardIntegrators(G);
  );
  
  Net->exampleHistoryStart = 0;

  Net->resetHistory[HISTORY_INDEX(0)] = TRUE;

  RUN_PROC(preEventProc);
  *correct = TRUE;
  for (; tick < Net->maxTicks && event < E->numEvents; tick++) {
    Net->currentTick = tick;
    Net->eventHistory[tick] = event;
    if (tick > 0) Net->resetHistory[HISTORY_INDEX(tick)] = FALSE;
    
    RUN_PROC(preTickProc);

    if( tick < mp )
      /* Next tick of forward pass on the input side, with no backprop */
      inputSideForward( (tick == mp - 1) );
    else
      /* Run the output side for the current tick, and accumulate derivs on middle layer */
      outputSideForwardBackward();


    RUN_PROC(postTickProc);
    
    updateDisplays(ON_TICK);
    
    eventDone = FALSE;
    critReached = 2;
    ticksOnEvent++;
    /* The SMALL_VAL is to correct for possible floating point inaccuracy */
    timeOnEvent = ((real) ticksOnEvent / Net->ticksPerInterval) + SMALL_VAL;
    /* If we have done enough time on this event, go on */
    if (timeOnEvent >= maxTime || tick >= Net->maxTicks - 1)
      eventDone = TRUE;
    /* If we have gone long enough and reached criterion, go on */
    else if (timeOnEvent >= minTime && 
	     (critReached = groupCriteriaReached(training)))
      eventDone = TRUE;
    /* See if the grace period is over */
    else if (Net->inGracePeriod && timeOnEvent >= graceTime)
      Net->inGracePeriod = FALSE;

    if (eventDone) {
      if (critReached == 2) critReached = groupCriteriaReached(training);
      if (!critReached) *correct = FALSE;
      RUN_PROC(postEventProc);
      updateDisplays(ON_EVENT);

      if (Net->groupCritRequired && !critReached) break;

      if (++event < E->numEvents && ((V = E->event + event))) {
	if (E->set->loadEvent(V)) return TCL_ERROR;
	ticksOnEvent = 0;
	maxTime = chooseValue(V->maxTime, E->set->maxTime);
	minTime = chooseValue(V->minTime, E->set->minTime);
	graceTime = chooseValue(V->graceTime, E->set->graceTime);
	Net->inGracePeriod = (graceTime > 0.0);
	RUN_PROC(preEventProc);
      }
    }
  }

  Net->ticksOnExample = Net->currentTick + 1;

  updateDisplays(ON_EXAMPLE);


  /* Backprop the whole input side for the last tick of the input side */
  int i, g; Group G, G1, G2;

  tick = mp - 1;
  for( g = 0; g < Net->ext->middleGroupNum; g++ ) {
    G = Net->group[g];
    restoreInputs( G, tick );
    restoreOutputs( G, tick );
  }

  inputSideBackward();  

  /* Continue to backprop in time to the first tick of the input side */

  G1 = Net->group[1]; G2 = Net->group[2];
  for( tick = E->numEvents / 2 - 2; tick >= 0; tick-- ) {
    /* assume that elman group is number 1, and copy output derivs to hidden, assumed to be group 2 */
    for( i = 0; i < G1->numUnits; i++ ) G2->unit[i].outputDeriv = G1->unit[i].outputDeriv;
    
    /* restore the inputs and outputs for this tick */
    for( g = 0; g < Net->ext->middleGroupNum; g++ ) {
      G = Net->group[g];
      restoreInputs( G, tick );
      restoreOutputs( G, tick );
      if( g < 2 ) resetOutputDerivs( G );
    }

    /* backprop from group 2, assumed to be the hidden group with incoming connections from elman */
    for( g = 2; g >= 0; g-- ) {
      G = Net->group[g];
      computeOutputBack(G);
      computeInputBack(G);
    }
  }

  return TCL_OK;
}


